/* General */

h1,h2,.rst-content .toctree-wrapper p.caption,h3,h4,h5,h6,legend {
  font-family: "Lato", "proxima-nova", "Helvetica Neue", Arial, sans-serif !important;
}

.rst-content span.highlighted {
  background: #FFC;
}

.btn-fat {
  padding: 1em !important;
}

.btn {
  border: 1px solid #404040;
}

.btn:focus {
  outline: 1px dashed !important;
}

:focus {
  outline: 1px dashed;
}
.wy-menu :not(.toctree-expand):focus {
  outline-offset: -1px;
}

#search-results .context {
  color: #404040 !important;
}

a {
  color: #2678AF;
}

.wy-menu-vertical li.current {
  border-left: 2px solid;
}

.wy-menu-vertical li.current a {
  color: #656565;
}

.wy-menu-vertical li.toctree-l2.current > a,
.wy-menu-vertical li.toctree-l2.current li.toctree-l3 > a {
  color: #555;
  z-index: 1;
}

.wy-menu-vertical li.toctree-l3.current li.toctree-l4 > a {
  color: #4b4b4b;
}

.wy-menu-vertical li button.toctree-expand {
  color: #b3b3b3;
}
.wy-menu-vertical li.toctree-l2 button.toctree-expand {
  color: #656565;
}

.wy-table-odd a, .wy-table-striped tr:nth-child(2n-1) a, .rst-content table.docutils:not(.field-list) tr:nth-child(2n-1) a {
  color: #2575AB;
}

.rst-content tt.literal, .rst-content tt.literal, .rst-content code.literal {
  color: #606060;
}

.wy-alert.wy-alert-info .wy-alert-title, .rst-content .note .wy-alert-title, .rst-content .wy-alert-info.attention .wy-alert-title, .rst-content .wy-alert-info.caution .wy-alert-title, .rst-content .wy-alert-info.danger .wy-alert-title, .rst-content .wy-alert-info.error .wy-alert-title, .rst-content .wy-alert-info.hint .wy-alert-title, .rst-content .wy-alert-info.important .wy-alert-title, .rst-content .wy-alert-info.tip .wy-alert-title, .rst-content .wy-alert-info.warning .wy-alert-title, .rst-content .seealso .wy-alert-title, .rst-content .wy-alert-info.admonition-todo .wy-alert-title, .wy-alert.wy-alert-info .rst-content .admonition-title, .rst-content .wy-alert.wy-alert-info .admonition-title, .rst-content .note .admonition-title, .rst-content .wy-alert-info.attention .admonition-title, .rst-content .wy-alert-info.caution .admonition-title, .rst-content .wy-alert-info.danger .admonition-title, .rst-content .wy-alert-info.error .admonition-title, .rst-content .wy-alert-info.hint .admonition-title, .rst-content .wy-alert-info.important .admonition-title, .rst-content .wy-alert-info.tip .admonition-title, .rst-content .wy-alert-info.warning .admonition-title, .rst-content .seealso .admonition-title, .rst-content .wy-alert-info.admonition-todo .admonition-title {
  background: #277BB3;
}

.wy-alert.wy-alert-warning .wy-alert-title, .rst-content .wy-alert-warning.note .wy-alert-title, .rst-content .attention .wy-alert-title, .rst-content .caution .wy-alert-title, .rst-content .wy-alert-warning.danger .wy-alert-title, .rst-content .wy-alert-warning.error .wy-alert-title, .rst-content .wy-alert-warning.hint .wy-alert-title, .rst-content .wy-alert-warning.important .wy-alert-title, .rst-content .wy-alert-warning.tip .wy-alert-title, .rst-content .warning .wy-alert-title, .rst-content .wy-alert-warning.seealso .wy-alert-title, .rst-content .admonition-todo .wy-alert-title, .wy-alert.wy-alert-warning .rst-content .admonition-title, .rst-content .wy-alert.wy-alert-warning .admonition-title, .rst-content .wy-alert-warning.note .admonition-title, .rst-content .attention .admonition-title, .rst-content .caution .admonition-title, .rst-content .wy-alert-warning.danger .admonition-title, .rst-content .wy-alert-warning.error .admonition-title, .rst-content .wy-alert-warning.hint .admonition-title, .rst-content .wy-alert-warning.important .admonition-title, .rst-content .wy-alert-warning.tip .admonition-title, .rst-content .warning .admonition-title, .rst-content .wy-alert-warning.seealso .admonition-title, .rst-content .admonition-todo .admonition-title {
  background: #B25E14;
}

div.admonition.note a {
  color: #2473A7;
}

/* version picker */

.rst-versions .rst-other-versions {
  color: #d9d9d9 !important;
}

.rst-other-versions a {
  color: #fcfcfc !important;
}

/* code blocks */

.highlight .c1, .highlight .cm {
  color: #3F7D8C !important;
}

.highlight .gp {
  color: #BD5A08 !important;
}

.highlight .nc, .highlight .nn {
  color: #0D7EAF !important;
}

.highlight .no {
  color: #2C7DA9 !important;
}

.highlight .nv {
  color: #B045CD !important;
}

.highlight .si {
  color: #3B7AB8 !important;
}

/* Side menu customization */

.wy-menu-vertical header, .wy-menu-vertical p.caption {
  color: #eee;
}

.wy-side-nav-search {
  background-color: #2473A7;
}

.wy-side-nav-search > div.version,
.wy-side-nav-search > a.icon-home,
a.skiplink {
  color: white !important;
}

/* Footer */

footer {
  color: #404040 !important;
}

footer span.commit code {
  color: #404040 !important;
}
